Package: eks
Version: 1.1.3
Date: 2026-05-10
Title: Tidy and Geospatial Kernel Smoothing
Authors@R: c(person("Tarn", "Duong", role=c("aut","cre"), email="tarn.duong@gmail.com", comment=c(ORCID="0000-0002-1198-3482"))) 
Maintainer: Tarn Duong <tarn.duong@gmail.com>
Depends: R (>= 2.10.0)
Imports: colorspace, dplyr, ggplot2 (>= 3.0.0), isoband, ks (>=
        1.15.0), sf
Suggests: ggquiver, knitr, mapsf, MASS, rmarkdown
VignetteBuilder: knitr
Encoding: UTF-8
Description: Extensions of the kernel smoothing functions from the 'ks' package for compatibility with the tidyverse and geospatial ecosystems <doi:10.1007/s00180-024-01543-9>.
License: GPL-2 | GPL-3
URL: https://mvstat.net/eks/, https://mvstat.net/mvksa/,
        https://github.com/cran/eks/
NeedsCompilation: no
Packaged: 2026-05-10 10:48:15 UTC; tduong
Author: Tarn Duong [aut, cre] (ORCID: <https://orcid.org/0000-0002-1198-3482>)
Repository: CRAN
Date/Publication: 2026-05-10 11:40:03 UTC
Built: R 4.6.0; ; 2026-05-10 13:25:39 UTC; unix
