importFrom(colorspace, diverging_hcl, qualitative_hcl, scale_colour_discrete_sequential, scale_colour_discrete_qualitative, scale_fill_discrete_diverging, sequential_hcl)
importFrom(dplyr, .data, arrange, as_tibble, enquo, group_by, group_indices, filter, group_modify, groups, left_join, mutate, one_of, pull, relocate, rename, select, slice_head, summarise, sym, tibble, ungroup, vars)
importFrom(ggplot2, ggplot, Geom, GeomContour, GeomContourFilled, GeomPoint, GeomRug, StatContour, StatContourFilled, aes, ggplot_add, guide_legend, guides, labs, scale_colour_manual, scale_colour_viridis_d, scale_fill_manual, scale_fill_viridis_d, standardise_aes_names, unit)
importFrom(geos, geos_area, geos_length, geos_intersects)
importFrom(graphics, image)
importFrom(grDevices, contourLines)
importFrom(isoband, isobands, isolines)
importFrom(ks, as.kde, contourBreaks, contourLevels, kcde, kcopula, kcurv, kda, kdcde, kde, kde.balloon, kde.boundary, kde.local.test, kde.sp, kde.test, kde.truncate, kdde.truncate, kdde, kdr, kfe, kfs, kms, kms.part, kroc, ksupp)
importFrom(lwgeom, st_split)
importFrom(sf, st_cast, st_as_sfc, st_as_sf, st_coordinates, st_crs, st_difference, st_drop_geometry, st_geometry, st_is, st_set_crs, st_sf, st_sfc)
importFrom(stats, cutree, dist, hclust, median, na.omit, predict, setNames, weighted.mean)
importFrom(tools, toTitleCase)
importFrom(utils, data, head, tail)

export(add_contour_breaks)
export(aes_ks, guides_ks, labs_ks)
export(as_tidy_ks)
export(dim_ks)
export(contour_breaks)
export(geom_contour_filled_ks, stat_contour_filled_ks)
export(geom_contour_ks, stat_contour_ks)            
export(geom_point_ks, stat_point_ks)
export(geom_rug_ks, stat_rug_ks)
export(GeomContourKs, StatContourKs)
export(GeomContourFilledKs, StatContourFilledKs)
export(GeomPointKs, StatPointKs)
export(GeomRugKs, StatRugKs)
export(label_percent)
export(scale_asymmetric) 
export(scale_transp_neutral)
export(st_add_coordinates)
export(st_get_contour)
export(st_remove_long_segment)
export(tidy_as_kde, st_as_kde)
export(tidy_intergrid, st_intergrid)
export(tidy_kcde, st_kcde)
export(tidy_kcopula)
export(tidy_kcurv, st_kcurv)
export(tidy_kda, st_kda)
export(tidy_kdcde, st_kdcde)
export(tidy_kdde, st_kdde)    
export(tidy_kde, st_kde)
export(tidy_kde_balloon, st_kde_balloon)   
export(tidy_kde_boundary, st_kde_boundary)
export(tidy_kde_local_test, st_kde_local_test)
export(tidy_kde_sp, st_kde_sp)
export(tidy_kde_truncate, st_kde_truncate)       
export(tidy_kdr, st_kdr)            
export(tidy_kfs, st_kfs)           
export(tidy_kms, st_kms)
export(tidy_kquiver, st_kquiver)
export(tidy_kroc, st_kroc)
export(tidy_ks, st_ks)           
export(tidy_ksupp, st_ksupp)
export(unfactor)  

S3method(c, sf_ks)
S3method(c, tidy_ks)
S3method(contourLevels, tidy_ks)
S3method(ggplot, sf_ks)
S3method(ggplot, tidy_ks)
S3method(ggplot_add, scale_rem)
S3method(plot, sf_ks)
S3method(rbind, sf_ks)
S3method(rbind, tidy_ks)
